/* GtkEditor - a source editor widget for GTK
 * Copyright (C) 1998, 2000 Thomas Mailund.
 *
 * The editor widget was written by Thomas Mailund, so bugs should be
 * reported to <mailund@daimi.au.dk>, not the gtk ppl.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>

#include <glib.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtkmain.h>
#include <gtk/gtksignal.h>

#include "gtkeditor.h"
#include "internal.h"
#include "tktext-port/gtktextbtree.h"


/* widget stuff */
static void   gtk_editor_class_init          (GtkEditorClass   *klass);
static void   gtk_editor_init                (GtkEditor        *editor);
static void   gtk_editor_destroy             (GtkObject        *object);


/* --<local data>--------------------------------------------------------- */
static GtkWidgetClass *parent_class = NULL;

/* --<widget initialization, constructor and destructor>------------------ */
guint
gtk_editor_get_type ()
{
  static guint editor_type = 0;

  if (!editor_type)
    {
      GtkTypeInfo editor_info =
      {
	"GtkEditor",
	sizeof (GtkEditor),
	sizeof (GtkEditorClass),
	(GtkClassInitFunc) gtk_editor_class_init,
	(GtkObjectInitFunc) gtk_editor_init,
	(GtkArgSetFunc) NULL,
        (GtkArgGetFunc) NULL,
      };

      editor_type = gtk_type_unique (gtk_text_buffer_get_type (), &editor_info);
    }

  return editor_type;
}

static void
gtk_editor_class_init (GtkEditorClass *class)
{
  GtkObjectClass *object_class;

  object_class = (GtkObjectClass*) class;

  /* init local data */
  parent_class = gtk_type_class (gtk_text_buffer_get_type ());


  /* setup signals */
  object_class->destroy = gtk_editor_destroy;
}

static void
gtk_editor_init (GtkEditor *editor)
{
  /* hilite */
  editor->scanner = NULL;
  editor->n_tag_entries = 0;
  editor->tag_entries = NULL;
  editor->tokens = NULL;
  editor->cur_token = NULL;
  editor->cur_token_pos = 0;
  editor->block_stack = NULL;

  gtk_signal_connect_after (GTK_OBJECT(editor),
                            "insert_text",
                            GTK_SIGNAL_FUNC(gtk_editor_insert_text),NULL);
  gtk_signal_connect (GTK_OBJECT(editor),
                            "delete_text",
                            GTK_SIGNAL_FUNC(gtk_editor_delete_text_length),NULL);
  gtk_signal_connect_after (GTK_OBJECT(editor),
                            "delete_text",
                            GTK_SIGNAL_FUNC(gtk_editor_delete_text),NULL);
}

GtkEditor*
gtk_editor_new ()
{
  GtkEditor *editor;

  editor = GTK_EDITOR (gtk_type_new (gtk_editor_get_type ()));
  editor->buffer.tag_table = gtk_text_tag_table_new();
  editor->buffer.tree = gtk_text_btree_new(editor->buffer.tag_table,
                                           &editor->buffer);
  editor->default_tag = gtk_text_buffer_create_tag(GTK_TEXT_BUFFER(editor), "default");

  return editor;
}

static void
gtk_editor_destroy (GtkObject *object)
{
  g_return_if_fail (object != NULL);
  g_return_if_fail (GTK_IS_EDITOR (object));

  GTK_OBJECT_CLASS(parent_class)->destroy (object);
}
