/* start condition definitions */
%x STRING COMMENT

/* pattern definitions */
D			[0-9]
L			[a-zA-Z_]
H			[a-fA-F0-9]
E			[Ee][+-]?{D}+
FS			(f|F|l|L)
IS			(u|U|l|L)*

keyword                 (break|case|continue|default|do|else|enum|for|goto|if|return|sizeof|struct|switch|typedef|union|while)
	
%%
 /* patterns */
"/*"           BEGIN_BLOCK("comment", 0, FALSE); BEGIN(COMMENT);
\"             BEGIN_BLOCK("string", 0, FALSE); BEGIN(STRING);

<COMMENT>{
  \n           TOKEN("new_line", 0);
  "*/"         END_BLOCK("comment"); BEGIN(INITIAL);
  .
}


<STRING>{
  \\\n         /* continued string */
  \n           END_BLOCK("string"); BEGIN(INITIAL);
  \\.          |
  \\[0-7]{1,3} TOKEN("escape_char", 1);
  \"           END_BLOCK("string"); BEGIN(INITIAL);
  .
}


^[ \t]*#[ \t]*include[ \t]*((\"[^<>\n]*\"?)|("<"[^\"\n]*">"?)) TOKEN("include", 0);

^[ \t]*#[ \t]*(include|if|ifdef|ifndef|undef|line|else|define|endif)  TOKEN("pre_processor", 0);

"auto"       |
"char"       |
"const"      |
"double"     |
"extern"     |
"float"      |
"int"        |
"long"       |
"register"   |
"short"      |
"signed"     |
"static"     |
"unsigned"   |
"void"       |
"volatile"   TOKEN("type", 0);

{keyword}                TOKEN("keyword", 0);
{keyword}/[ \t\n]*\(     TOKEN("keyword", 0);
	
{L}({L}|{D})*/[ \t\n]*\( TOKEN("function",0); 

{L}({L}|{D})*		/* identifier ignore */

0[xX]{H}+{IS}?		|
0{D}+{IS}?		|
{D}+{IS}?		|
'(\\.|[^\\'])+'		|

{D}+{E}{FS}?		|
{D}*"."{D}+({E})?{FS}?	|
{D}+"."{D}*({E})?{FS}?	TOKEN("constant", 0);	

">>=" |
"<<=" |
"+="  |
"-="  |
"*="  |
"/="  |
"%="  |
"&="  |
"^="  |
"|="  |
">>"  |
"<<"  |
"++"  |
"--"  |
"->"  |
"&&"  |
"||"  |
"<="  |
">="  |
"=="  |
"!="  |
"="   |
"&"   |
"!"   |
"~"   |
"-"   |
"+"   |
"*"   |
"/"   |
"%"   |
"<"   |
">"   |
"^"   |
"|"   |
"?"   TOKEN("operator", 0);

";"
"{"
"}"
","
":"

"("
")"
"["
"]"
"."


[ \t\v\n\f]		
.	     /* ignore bad characters */
