/* start condition definitions */
%x STRING
%x FUNC

/* pattern definitions */
D			[0-9]
L			[a-zA-Z_]
H			[a-fA-F0-9]
E			[Ee][+-]?{D}+
FS			(f|F|l|L)
IS			(u|U|l|L)*
KEYWORD                 (=>|and|case|cond|define|delay|do|else|if|lambda|let|let*|letrec|or|quasiquote|quote|set!|unquote|unquaote-splicing)

%%
 /* patterns */
";"            BEGIN_BLOCK("comment", 0, FALSE);
"\n"           END_BLOCK("comment");
\"             BEGIN_BLOCK("string", 0, FALSE); BEGIN(STRING);

<STRING>{
  \\\n         /* continued string */
  \n           END_BLOCK("string"); BEGIN(INITIAL);
  \\.          |
  \\[0-7]{1,3} TOKEN("escape_char", 0);
  \"           END_BLOCK("string"); BEGIN(INITIAL);
  .
}

{KEYWORD}           TOKEN("keyword", 0);

"("            BEGIN(FUNC);

<FUNC>{
  {KEYWORD}         TOKEN("keyword", 0);
  [^ \t\v\n\f()]*   TOKEN("function", 0); BEGIN(INITIAL);
  .                 BEGIN(INITIAL);
}

{L}({L}|{D})*		/* identifier ignore */

0[xX]{H}+{IS}?		|
0{D}+{IS}?		|
{D}+{IS}?		|
'(\\.|[^\\'])+'		|

{D}+{E}{FS}?		|
{D}*"."{D}+({E})?{FS}?	|
{D}+"."{D}*({E})?{FS}?	TOKEN("constant", 0);	

[ \t\v\n\f]		
.	     /* ignore bad characters */
