/* start condition definitions */
%x STRING

/* pattern definitions */
D			[0-9]
L			[a-zA-Z_]
H			[a-fA-F0-9]
E			[Ee][+-]?{D}+
FS			(f|F|l|L)
IS			(u|U|l|L)*

%%
 /* patterns */
"(*"           BEGIN_BLOCK("comment", 0, TRUE);
"*)"           END_BLOCK("comment");
\"             BEGIN_BLOCK("string", 0, FALSE); BEGIN(STRING);

<STRING>{
  \\\n         /* continued string */
  \n           END_BLOCK("string"); BEGIN(INITIAL);
  \\.          |
  \\[0-7]{1,3} TOKEN("escape_char", 0);
  \"           END_BLOCK("string"); BEGIN(INITIAL);
  .
}


"char"       |
"real"       |
"int"        |
"string"     |
"bool"       TOKEN("type", 0);

"if"         |
"then"       |
"else"       |
"case"       |
"of"         |
"|"          |
"=>"         |
"fn"         |
"fun"        |
"let"        |
"in"         |
"end"        |
"while"      TOKEN("keyword", 0);

{L}({L}|{D})*		/* identifier ignore */

0[xX]{H}+{IS}?		|
0{D}+{IS}?		|
{D}+{IS}?		|
'(\\.|[^\\'])+'		|

{D}+{E}{FS}?		|
{D}*"."{D}+({E})?{FS}?	|
{D}+"."{D}*({E})?{FS}?	TOKEN("constant", 0);	

"!"   |
"~"   |
"-"   |
"+"   |
"*"   |
"/"   |
"%"   |
"<"   |
">"   |
"^"   |
"="   TOKEN("operator", 0);

";"
"{"
"}"
","
":"

"("
")"
"["
"]"
"."


[ \t\v\n\f]		
.	     /* ignore bad characters */
