/* main.l - scanner for gtkeditor language input */

%{

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
		
#define OUTPUT fwrite(yytext, yyleng, 1, output_file)

#define FALSE 0
#define TRUE 1
#define MAXLINE 1024

static FILE *output_file;
static int sectnum = 0;
char *scanner_name = NULL;

int get_token_id(char *name);
int get_block_id(char *name);

%}

%option caseless nodefault outfile="main.c" stack noyy_top_state
%option nostdinit

%x SECT2 SECT2PROLOG SECT3 CODEBLOCK PICKUPDEF SC CARETISBOL NUM QUOTE
%x FIRSTCCL CCL ACTION RECOVER COMMENT ACTION_STRING PERCENT_BRACE_ACTION
%x OPTION LINEDIR NAME_STRING PRECEDENCE NESTING TOKEN_NAME BLOCK_NAME
%x END_TOKEN END_BLOCK_BEGIN END_BLOCK_END PATTERN SC_BLOCK  PATTERN_STRING

WS		[[:blank:]]+
OPTWS		[[:blank:]]*
NOT_WS		[^[:blank:]\n]

NL		\r?\n

NAME		([[:alpha:]_][[:alnum:]_-]*)
NOT_NAME	[^[:alpha:]_*\n]+

SCNAME		{NAME}

ESCSEQ		(\\([^\n]|[0-7]{1,3}|x[[:xdigit:]]{1,2}))

FIRST_CCL_CHAR	([^\\\n]|{ESCSEQ})
CCL_CHAR	([^\\\n\]]|{ESCSEQ})
CCL_EXPR	("[:"[[:alpha:]]+":]")

LEXOPT		[aceknopr]

%%
	static int bracelevel, indented_code;
	static int doing_rule_action = FALSE;
	static int linenum = 0;

	int doing_codeblock = FALSE;
        int token_id, block_id;

	char buffer[MAXLINE], precedence[MAXLINE], nesting[MAXLINE], *string_ptr, prefix[4];


<INITIAL>{
	^{WS}		indented_code = TRUE; BEGIN(CODEBLOCK);
	^"%{".*{NL}	{
	                OUTPUT;
			++linenum;
			indented_code = FALSE;
			BEGIN(CODEBLOCK);
			}

	^"%%".*		{
	                srand(time(NULL));
	                prefix[0] = 97 + (int)(26.0*rand()/(RAND_MAX+1.0));
	                prefix[1] = 97 + (int)(26.0*rand()/(RAND_MAX+1.0));
	                prefix[2] = 97 + (int)(26.0*rand()/(RAND_MAX+1.0));
	                prefix[3] = '\0';
	                fprintf (output_file, "%%option fast always-interactive noyywrap prefix=\"%s\"\n", prefix);
	                fprintf (output_file, "%%option noyy_scan_buffer noyy_scan_bytes noyy_scan_string\n\n");
	                fprintf (output_file, "%%{\n\n");
	                fprintf (output_file, "#define FALSE 0\n");
	                fprintf (output_file, "#define TRUE 1\n\n");
                        fprintf (output_file, "#define YY_INPUT(buf,result,max_size) \\\n");
                        fprintf (output_file, "{ \\\n");
                        fprintf (output_file, "  int c = next_char(); \\\n");
                        fprintf (output_file, "  result = (c == EOF) ? YY_NULL : (buf[0] = c, 1); \\\n");
                        fprintf (output_file, "}\n\n");
                        fprintf (output_file, "#define YY_DECL int lex ()\n");
                        fprintf (output_file, "#define YY_USER_ACTION position += prev_leng; prev_leng = yyleng;\n\n");
                        /* definitions */
	                fprintf (output_file, "\nstatic void (*new_token)(int id, int precedence, char *text, int length, int position, int block_id, int nesting, int block_type);\n");
	                fprintf (output_file, "\nstatic int (*next_char)();\n\n");
	                fprintf (output_file, "%%}\n\n");
	                OUTPUT;
	                fprintf (output_file, "\n\n%%{\nint position = 0, prev_leng = 0;\n%%}\n\n");
			sectnum = 2;
			bracelevel = 0;
			BEGIN(SECT2PROLOG);
			}

	.	        OUTPUT;
	^{OPTWS}{NL}	OUTPUT; ++linenum; /* allows blank lines in section 1 */
	{OPTWS}{NL}	OUTPUT; ++linenum; /* maybe end of comment line */
}


<COMMENT>{
	"*/"		OUTPUT; yy_pop_state();
	"*"		OUTPUT;
	[^*\n]+		OUTPUT;
	[^*\n]*{NL}	OUTPUT; ++linenum;
}

<CODEBLOCK>{
	^"%}".*{NL}	 {
	                   ++linenum;
	                   BEGIN(INITIAL);

	                   OUTPUT;
	                 }

	{NAME}|{NOT_NAME}|.	  OUTPUT;

	{NL}		{
	                OUTPUT;
			++linenum;
			if ( indented_code )
				BEGIN(INITIAL);
			}
}


<SECT2PROLOG>{
	^"%{".*	OUTPUT; ++bracelevel;
	^"%}".*	OUTPUT; --bracelevel;

	^{WS}.*	OUTPUT;	/* indented code in prolog */

	^{NOT_WS}.*	{	/* non-indented code */
			if ( bracelevel <= 0 )
				{ /* not in %{ ... %} */
				yyless( 0 );	/* put it all back */
				yy_set_bol( 1 );
				BEGIN(SECT2);
				}
			}

	.*		OUTPUT;
	{NL}	        OUTPUT; ++linenum;
}

<SECT2>{
	{OPTWS}{NL}	OUTPUT; ++linenum; /* allow blank lines in section 2 */

	^{OPTWS}"%{"	{
	                OUTPUT;
			bracelevel = 1;
			BEGIN(PERCENT_BRACE_ACTION);
			}
			
	^"<<EOF>>"      OUTPUT; yy_push_state(PATTERN);
	^"<"            OUTPUT; yy_push_state(SC);
	
	^{NOT_WS}       yyless(0); yy_push_state(PATTERN);


	^{WS}"/*"	{
	                OUTPUT;
			yy_push_state(COMMENT);
			}


	^"%%".*		{
	                OUTPUT;
			sectnum = 3;
			BEGIN(SECT3);
			}

	.|{NL}		OUTPUT;
}

<PATTERN>{
  "<<EOF>>"             OUTPUT;
  \"                    OUTPUT; yy_push_state(PATTERN_STRING);
  \\.                   OUTPUT;
  \[                    OUTPUT; yy_push_state(CCL);
  {WS}                  OUTPUT; yy_pop_state(); yy_push_state(ACTION);
  .                     OUTPUT;
  \n                    OUTPUT; yy_pop_state();
}

<PATTERN_STRING>{
\\.      OUTPUT;
\"       OUTPUT; yy_pop_state();
.        OUTPUT;
}


<SC>{
	[,*]		OUTPUT;
	">"		OUTPUT; yy_push_state(PATTERN);
	">"[\t\n ]*\{	OUTPUT; yy_push_state(SC_BLOCK);
	{SCNAME}	OUTPUT;
	{WS}            OUTPUT;
}

<SC_BLOCK>{
	{OPTWS}{NL}	OUTPUT; ++linenum; /* allow blank lines in section 2 */

	^{OPTWS}"%{"	{
	                OUTPUT;
			bracelevel = 1;
			BEGIN(PERCENT_BRACE_ACTION);
			}
			
	^{OPTWS}"<<EOF>>"  OUTPUT; yy_push_state(PATTERN);

	^{OPTWS}"<"            OUTPUT; yy_push_state(SC);
	
	^{OPTWS}{NOT_WS}  yyless(yyleng-1); yy_push_state(PATTERN);


	^{WS}"/*"	{
	                OUTPUT;
			yy_push_state(COMMENT);
			}

	\}		OUTPUT; yy_pop_state(); yy_pop_state();
}

<CCL>{
        "["             OUTPUT; yy_push_state(CCL);
	"]"		OUTPUT; yy_pop_state();
	\\\[             OUTPUT;
	\\\]             OUTPUT;
	.		OUTPUT;
}



<PERCENT_BRACE_ACTION>{
	{OPTWS}"%}".*		OUTPUT; bracelevel = 0;

	<ACTION>"/*"		OUTPUT; yy_push_state( COMMENT );

	<CODEBLOCK,ACTION>{
		"reject"	{
			OUTPUT;
			}
		"yymore"	{
		        OUTPUT;
			}
	}

	{NAME}|{NOT_NAME}|.	OUTPUT;
	{NL}		{
	                OUTPUT;
			++linenum;
			if ( bracelevel == 0 ||
			     (doing_codeblock && indented_code) )
				{
				doing_rule_action = doing_codeblock = FALSE;
				BEGIN(SECT2);
				
				}
			}
			
	TOKEN[\n \t]*\(  {
	                fprintf(output_file, "new_token (");
	                yy_push_state(END_TOKEN);
	                yy_push_state(PRECEDENCE);
	                yy_push_state(TOKEN_NAME);
	                }
	
	BEGIN_BLOCK[\n \t]*\(  {
	                fprintf(output_file, "new_token (");
	                yy_push_state(END_BLOCK_BEGIN);
	                yy_push_state(TOKEN_NAME);
	                yy_push_state(NESTING);
	                yy_push_state(PRECEDENCE);
	                yy_push_state(BLOCK_NAME);
	                }
	END_BLOCK[\n \t]*\(  {
	                fprintf(output_file, "new_token (");
	                yy_push_state(END_BLOCK_END);
	                yy_push_state(TOKEN_NAME);
	                yy_push_state(BLOCK_NAME);
	                token_id = -1;
	                }
}


	/* Reject and YYmore() are checked for above, in PERCENT_BRACE_ACTION */
<ACTION>{
        TOKEN[\n \t]*\(  {
	                fprintf(output_file, "new_token (");
	                yy_push_state(END_TOKEN);
	                yy_push_state(PRECEDENCE);
	                yy_push_state(TOKEN_NAME);
	                }
	
	BEGIN_BLOCK[\n \t]*\(  {
	                fprintf(output_file, "new_token (");
	                yy_push_state(END_BLOCK_BEGIN);
	                yy_push_state(TOKEN_NAME);
	                yy_push_state(NESTING);
	                yy_push_state(PRECEDENCE);
	                yy_push_state(BLOCK_NAME);
	                }
	END_BLOCK[\n \t]*\(  {
	                fprintf(output_file, "new_token (");
	                yy_push_state(END_BLOCK_END);
	                yy_push_state(TOKEN_NAME);
	                yy_push_state(BLOCK_NAME);
	                token_id = -1;
	                }
	
	"{"		OUTPUT; ++bracelevel;
	"}"		OUTPUT; --bracelevel;


	"'"([^'\\\n]|\\.)*"'"	OUTPUT; /* character constant */
	\"		OUTPUT; BEGIN(ACTION_STRING);
	{NL}		{
			++linenum;
			
			if ( bracelevel == 0 )
				{
				doing_rule_action = FALSE;
				yy_pop_state();
				}
			OUTPUT;
			}
	
	
	.		OUTPUT;
}

<TOKEN_NAME>{
{WS}

\"          string_ptr = buffer; yy_push_state(NAME_STRING);

,         {
            token_id = get_token_id(buffer);
            yy_pop_state();
          }
\)        {
            token_id = get_token_id(buffer);
            yyless(0);
            yy_pop_state();
          }
.         printf("token names must be constant");
}

<PRECEDENCE>{
[^,)]+      strcpy(precedence, yytext);
,            yy_pop_state();
\)          yyless(0); yy_pop_state();
}



<END_TOKEN>{
{WS}
\)     {
         fprintf(output_file, "%d, %s, yytext, yyleng, position, -1, 0, 0)",
            token_id, precedence);
         yy_pop_state();
       }
.      OUTPUT; printf("warning missing \")\""); yy_pop_state();
}

<BLOCK_NAME>{
{WS}

\"          string_ptr = buffer; yy_push_state(NAME_STRING);

,         {
            block_id = get_block_id(buffer);
            yy_pop_state();
          }
\)        {
            block_id = get_block_id(buffer);
            yyless(0);
            yy_pop_state();
            yy_pop_state();
          }
.         printf("block names must be constant");
}

<NESTING>{
{WS}
,           yy_pop_state();
\)          yy_pop_state(); yy_pop_state(); yyless(0); token_id = -1;
[^,)]+      strcpy(nesting, yytext);
}

<END_BLOCK_BEGIN>{
{WS}
\)     {
         fprintf(output_file, "%d, %s, yytext, yyleng, position, %d, %s, 1)",
            token_id, precedence, block_id, nesting);
         yy_pop_state();
       }
.      printf("error");
}

<END_BLOCK_END>{
{WS}
\)     {
         fprintf(output_file, "%d, 0, yytext, yyleng, position, %d, 0, 0)",
            token_id, block_id);
         yy_pop_state();
       }
.      printf("error");
}




<NAME_STRING>{
        \"        {
                     yy_pop_state();
                     *string_ptr = '\0';
                  }

        \n        {
                     printf("unterminated block name");
                  }

        \\[0-7]{1,3} {
        /* octal escape sequence */
        int result;

        sscanf( yytext + 1, "%o", &result );

        if ( result > 0xff )
            printf("octal escape out of bounds");

        *string_ptr++ = result;
        }

        \\[0-9]+  printf("error - bad escape sequence");
\\n  *string_ptr++ = '\n';
\\t  *string_ptr++ = '\t';
\\r  *string_ptr++ = '\r';
\\b  *string_ptr++ = '\b';
\\f  *string_ptr++ = '\f';

\\(.|\n)  *string_ptr++ = yytext[1];

[^\\\n\"]+        {
        char *yptr = yytext;

        while ( *yptr )
                *string_ptr++ = *yptr++;
        }
}

<ACTION_STRING>{
	[^"\\\n]+	OUTPUT;
	\\.		OUTPUT;
	{NL}		OUTPUT; ++linenum;
	\"		OUTPUT; BEGIN(ACTION);
	.		OUTPUT;
}


<SECT3>{
	.*(\n?)		OUTPUT;
	<<EOF>>		OUTPUT; sectnum = 0; yyterminate();
}

<*>.|\n			OUTPUT; printf( "bad character: %s" , yytext);

%%

typedef struct _Token {
  char *name;
  struct _Token *next;
}Token;

static Token *token_list = NULL;

int get_token_id(char *name)
{
  Token *cur, *prev;
  int i = 0;

  for(cur = token_list; cur != NULL; cur = cur->next) {
    if(strcmp(cur->name, name) == 0) {
      return i;
    }
    prev = cur;
    i++;
  }

  if(token_list == NULL)
    cur = token_list = malloc(sizeof(Token));
  else
    cur = prev->next = malloc(sizeof(Token));
  cur->name = strdup(name);
  cur->next = NULL;
  return i;
}


typedef struct _Block {
  char *name;
  struct _Block *next;
}Block;

static Block *block_list = NULL;

int get_block_id(char *name)
{
  Block *cur, *prev;
  int i = 0;

  for(cur = block_list; cur != NULL; cur = cur->next) {
    if(strcmp(cur->name, name) == 0) {
      return i;
    }
    prev = cur;
    i++;
  }

  if(block_list == NULL)
    cur = block_list = malloc(sizeof(Block));
  else
    cur = prev->next = malloc(sizeof(Block));
  cur->name = strdup(name);
  return i;
}

void write_scanner_interface()
{
  Token *token;
  Block *block;
  int i;

  /* write the interface */

  /* get_name */
  fprintf (output_file, "\nconst char *\nget_name ()\n");
  fprintf (output_file, "\n{\n  const char *name = ");
  if (scanner_name)
    fprintf (output_file, "\"%s\";\n\n", scanner_name);
  else
    fprintf (output_file, "NULL;\n\n");
  fprintf (output_file, "  return name;\n}\n\n");

  /* get_token_names */
  fprintf (output_file, "\nconst char **\nget_token_names ()\n");
  fprintf (output_file, "\n{\n  static const char *token_names[] = ");
  if (token_list) {
    fprintf (output_file, "{");
    for(token = token_list; token->next != NULL; token = token->next) {
      fprintf (output_file, "\"%s\", ", token->name);
    }
    fprintf (output_file, "\"%s\", NULL};\n\n", token->name);
  }
  else
    fprintf (output_file, "{NULL};\n\n");
  fprintf (output_file, "  return token_names;\n}\n\n");

  /* get_block_names */
  fprintf (output_file, "const char **\nget_block_names ()\n");
  fprintf (output_file, "\n{\n  static const char *block_names[] = ");
  if(block_list) {
    fprintf (output_file, "{");
    for(block = block_list; block->next != NULL; block = block->next) {
      fprintf (output_file, "\"%s\", ", block->name);
    }
    fprintf (output_file, "\"%s\", NULL};\n\n", block->name);
  }
  else
    fprintf (output_file, "{NULL};\n");
  fprintf (output_file, "  return block_names;\n}\n\n");

  /* get_token_no */
  fprintf (output_file, "\nint\nget_token_no (const char *name)\n");
  fprintf (output_file, "\n{\n  int number = -1;\n\n");
  if (token_list) {
    fprintf (output_file, "  if (strcmp(\"%s\", name) == 0) {\n", token_list->name);
    fprintf (output_file, "    return 0;\n");
    i = 1;
    for(token = token_list->next; token != NULL; token = token->next) {
      fprintf (output_file, "  } else if (strcmp(\"%s\", name) == 0) {\n", token->name);
      fprintf (output_file, "    return %d;\n", i);
      i++;
    }
    fprintf (output_file, "  }\n\n");
  }

  fprintf (output_file, "  return number;\n}\n\n");

  /* get_block_no */
  fprintf (output_file, "\nint\nget_block_no (const char *name)\n");
  fprintf (output_file, "\n{\n  int number = -1;\n\n");
  if (block_list) {
    fprintf (output_file, "  if (strcmp(\"%s\", name) == 0) {\n", block_list->name);
    fprintf (output_file, "    return 0;\n");
    i = 1;
    for(block = block_list->next; block != NULL; block = block->next) {
      fprintf (output_file, "  } else if (strcmp(\"%s\", name) == 0) {\n", block->name);
      fprintf (output_file, "    return %d;\n", i);
      i++;
    }
    fprintf (output_file, "  }\n\n");
  }

  fprintf (output_file, "  return number;\n}\n\n");

  /* token_func */
  fprintf (output_file, "void\ntoken_func (void (*func)(int id, int precedence, char *text, int length, int position, int block_id, int nesting, int block_type))\n");
  fprintf (output_file, "{\n  new_token = func;\n}\n\n");

  /* char_func */
  fprintf (output_file, "void\nchar_func (int (*func)())\n");
  fprintf (output_file, "{\n  next_char = func;\n}\n\n");

  /* get_state */
  fprintf (output_file, "int\nget_state ()\n");
  fprintf (output_file, "{\n  return YY_START;\n}\n\n");

  /* set_state */
  fprintf (output_file, "void\nset_state (int state)\n");
  fprintf (output_file, "{\n  BEGIN (state);\n}");
}

int main(int argc, char **argv)
{
  FILE *input_file;
  YY_BUFFER_STATE buffer;
  char file_name[256];

  switch (argc) {
    case 2:
      sprintf(file_name, "scanner.l");
      break;
    case 3:
      strcpy(file_name, argv[2]);
      break;
    default:
      printf("Usage: scannertool input_file [output_file]");
      return 0;
  }


  if (!(input_file = fopen(argv[1], "r"))) {
     printf ("Can\'t open file %s", argv[1]);
     return 0;
  }
  output_file = fopen(file_name, "w");

  buffer = yy_create_buffer(input_file, YY_BUF_SIZE);
  yy_switch_to_buffer(buffer);
  printf("lexing ...\n");
  yylex();
  printf("done\n");
  yy_delete_buffer(buffer);
  fclose(input_file);

  if(sectnum == 2) {
    fprintf(output_file, "\n\n%%%%\n\n");
  }

  printf("writing the interface\n");
  write_scanner_interface();

  fclose(output_file);

  return 0;
}

int yywrap()
{
  return 1;
}